
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Scanner;
import java.util.TreeMap;

/*
 * This class stores information about various stocks.
 * The information includes the name of the company and 
 * the stock symbol along with stock fundamentals such as
 * change in price, dividend ratio, and price to earnings
 * raio. The names and symbols are read from files.
 * The data values for the for the stocks mat be read from
 * a fixed file or obtained from the web via Yahoo finance.
 */
public class Stocks {
    
    // class constants for various file names
    private static final String STOCK_SYMBOL_FILE = "company_names_and_symbols.txt";
    private static final String DATA_SYMBOL_FILE = "data_fields.txt";
    private static final String SAMPLE_FILE = "sample_stock_data.txt";

    // holds the stock data
    private ArrayList<StockData> stocks;
    
    // holds sorted versions of the stocks
    private TreeMap<String, List<SimpleStockData>> sortedVersions;
    private TreeMap<String, String> stockSymbolsAndNames;
    
    private boolean useFile;
    
    // Use Yahoo finance to pull basic stock data. (or use sample file)
    // Info on symbols and use from these web pages:
    // http://www.gummy-stuff.org/Yahoo-data.htm
    // http://benjisimon.blogspot.com/2009/01/truly-simple-stock-api.html
    public Stocks(boolean uf) {
        useFile = uf;
        stockSymbolsAndNames 
                = getSymbolsAndNames(STOCK_SYMBOL_FILE);
        if(useFile) {
            stocks = createStocks(SAMPLE_FILE, true);
            createdSortedVersions();
        }
        else {
            setStockDataFromWeb();
        }
        
        // for debugging uncomment the following line
        // printStocks();
    }
    
    
    public void updateStockData() {
        setStockDataFromWeb();
    }
    
    private void setStockDataFromWeb() {
        // base url to get data
        String baseURL = "http://finance.yahoo.com/d/quotes.csv?s=";

        // add the stocks we want data for
        baseURL = addSymbols(baseURL, stockSymbolsAndNames.keySet(), "+");

        // after the stock symbols are added the symbols 
        // for the requested data fields are added
        baseURL += "&f=";
        baseURL = addSymbols(baseURL, readSymbols(DATA_SYMBOL_FILE), "");
        stocks = createStocks(baseURL, false);
        // showResult(baseURL);   
        
        createdSortedVersions();
    }
    
    //return array list of sorted data
    // format of string is "<Name>,<Symbol>,<Data>"
    public List<SimpleStockData> getSortedData(String dataType) {
        return sortedVersions.get(dataType);
    }

    
    
    // store sorted versions of arrays
    // to allow generalization and avoid repeated code
    // I store the method names from StockData and use
    // reflection to get the methods and then call them
    // on StockData objects. The assumption is all these
    // methods return doubles.
    private void createdSortedVersions() {
        // method getting long. How to break up?
        StockDataMethodInfo info = new StockDataMethodInfo();
        sortedVersions = new TreeMap<String, List<SimpleStockData>>();
        for(int i = 0; i < info.size(); i++) {
            Method methodForData = null;
            try {
                methodForData = StockData.class.getMethod(info.getMethodName(i));
            } catch (SecurityException e) {
                System.out.println("Problem trying to get method: " + e);
                e.printStackTrace();
            } catch (NoSuchMethodException e) {
                System.out.println("Problem trying to get method: " + e);
                e.printStackTrace();
            }
            Comparator<StockData> comp 
                = getComparator(methodForData, info.getSubFirst(i));
            
            List<SimpleStockData> sortedVersion 
                = getSortedVersion(comp, methodForData);
            
            sortedVersion = Collections.unmodifiableList(sortedVersion);
            sortedVersions.put(info.getDataName(i), sortedVersion);
        }
    }


    // DEBUG method to show Stocks, must create toString in StockData
    private void printStocks() {
        for(StockData stock : stocks)
            System.out.println(stock);
    }


    // create an ArrayList of StockData objects
    // url has already been built
    // I assume every stock requested in the URL is in symbolsAndNames
    
    // expected format of data is "<change> - <percent change>%",<averageVolume>,<volume>
    // <yield>,<P/E ratio>,<marketCap>B
    private ArrayList<StockData> createStocks(String dataSource, boolean isFile) {
        ArrayList<StockData> result = new ArrayList<StockData>();
        Scanner sc = createScanner(dataSource, isFile);
        while(sc.hasNextLine()) {
            String data = sc.nextLine();
            // trim off the symbol
            int indexLastComma = data.lastIndexOf(",");
            String symbol = data.substring(indexLastComma + 1);
            // check for quotes
            assert symbol.charAt(0) == '"' 
                && symbol.charAt(symbol.length() - 1) == '"' :
                    "Error in symbol format. Missing quotes at start and end: " + symbol;
            // trim quotes
            symbol = symbol.substring(1, symbol.length() - 1);
            
            // don't need symbol in raw data line so trim it off
            data = data.substring(0, indexLastComma);
            result.add(new StockData(stockSymbolsAndNames.get(symbol), symbol, data));
        }
        return result;
    }

    
    // create scanner hooked up to url or file as appopriate
    private Scanner createScanner(String dataSource, boolean isFile) {
        Scanner sc = null;
        try {  
            if(isFile) {
                sc =new Scanner(new File(dataSource));
            }
            else {
                URL stockPage = new URL(dataSource);
                sc =new Scanner(new InputStreamReader(stockPage.openStream()));
            } 
        }
        catch(IOException e) {
            System.out.println("ERROR reading from web page: " + e);
            System.out.println("no data available");
            e.printStackTrace();
        }
        return sc;
    }

    
    // assume file format is name,symbol
    // return map with symbol as key and name as value
    private TreeMap<String, String> getSymbolsAndNames(
            String fileName) {
        TreeMap<String, String> result = new TreeMap<String, String>();
        try {
            Scanner sc = new Scanner(new File(fileName));
            while(sc.hasNextLine()) {
                String line = sc.nextLine().trim();
                int indexOfComma = line.indexOf(",");
                String name = line.substring(0, indexOfComma);
                String symbol = line.substring(indexOfComma + 1);
                result.put(symbol, name);
            };
            sc.close();
        }
        catch(IOException e) {
            System.out.println("ERROR reading from file: " + e);
            System.out.println("returning emptying string for url");
            e.printStackTrace();
        }
        return result;
    }


    // DEBUG method that simply prints out results from URL
    private void showResult(String url) {
        System.out.println(url);
        try {
            URL stockPage = new URL(url);
            Scanner sc =new Scanner(new InputStreamReader(stockPage.openStream()));
            while(sc.hasNextLine())
                System.out.println(sc.nextLine());
        }
        catch(IOException e) {
            System.out.println("ERROR reading from web page: " + e);
            System.out.println("no data available");
        }
    }

    
    // Add the give values to the baseURL following each value by a seperator.
    // The last seperator is removed.
    private String addSymbols(String baseURL, Collection<String> values, String seperator) {
        StringBuilder result = new StringBuilder(baseURL);
        for(String symbol : values) {
            result.append(symbol);
            result.append(seperator);
        }
        // remove last seperator, unless empty string
        if(seperator.length() > 0) {
            result.delete(result.length() 
                    - seperator.length(), result.length());
        }
        return result.toString();
    }


    // read in the data symbols from a file
    private ArrayList<String> readSymbols(String fileName) {
        ArrayList<String> result = new ArrayList<String>();
        try {
            Scanner sc = new Scanner(new File(fileName));
            while(sc.hasNextLine()) {
                result.add(sc.nextLine().trim());
            };
            sc.close();
        }
        catch(IOException e) {
            System.out.println("ERROR reading from file: " + e);
            System.out.println("returning emptying string for url");
        }
        return result;
    }
    
    
    // return an ArrayList<String> in sorted order 
    // based on the method sent in.
    // the method must exist and is assumed to return a double.
    // sorted in descending order
    private ArrayList<SimpleStockData> getSortedVersion(Comparator<StockData> comp,
            Method methodForData) {
        Collections.sort(stocks, comp);
        ArrayList<SimpleStockData> result = new ArrayList<SimpleStockData>();
        for(StockData data : stocks) {
            double dataField = 0.0;
            try {
                dataField = (Double) methodForData.invoke(data);
            } catch(InvocationTargetException e) {
                System.out.println("problem calling method: " + e);
                e.printStackTrace();
            }   catch(IllegalAccessException e) {
                System.out.println("problem calling method: " + e);
                e.printStackTrace();
            }
            SimpleStockData temp = new SimpleStockData(
                    data.getName(), data.getSymbol(), dataField);
            result.add(temp);
        }
    
        return result;
    }
    
    

    
    
    // create compartor based on method passed in.
    // we don't override equals as these are private
    // and we won't allow access to the Comparators
    private Comparator<StockData> getComparator(final Method method,
                final boolean first) {
        return new Comparator<StockData>() {
            public int compare(StockData s1, StockData s2) {
                double diff = 0.0;
                try {
                    double resultS1 = (Double) method.invoke(s1);
                    double resultS2 = (Double) method.invoke(s2);
                    diff = first ? resultS1 - resultS2 : resultS2 - resultS1;
                } catch(InvocationTargetException e) {
                    System.out.println("problem calling method: " + e);
                    e.printStackTrace();
                } catch(IllegalAccessException e) {
                    System.out.println("problem calling method: " + e);
                    e.printStackTrace();
                }
                // want descending order so logic reverse of expected 
                return (diff > 0) ? -1 : (diff < 0) ? 1 : 
                    s1.getName().compareTo(s2.getName());
            }
        };
    }
 
    
//    // tester
//    public static void main(String[] args) {
//        Stocks sc = new Stocks(false);
//        List<SimpleStockData> test = sc.getSortedData("Percent Change");
//        for(SimpleStockData st : test)
//            System.out.println(st);
//    }
}
